@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004,2007 Choe Hwanjin
+ * Copyright (C) 2003-2008 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,8 +25,12 @@
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
+#include "debug.h"
 #include "server.h"
 #include "candidate.h"
+#include "gettext.h"
+#include "util.h"
+#include "nabi.h"
 
 enum {
     COLUMN_INDEX,
@@ -35,6 +39,20 @@ enum {
     NO_OF_COLUMNS
 };
 
+static void
+nabi_candidate_on_format(GtkWidget* widget, gpointer data)
+{
+    if (data != NULL &&
+	gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
+	if (strcmp(nabi->config->candidate_format, (const char*)data) != 0) {
+	    g_free(nabi->config->candidate_format);
+	    nabi->config->candidate_format = g_strdup((const char*)data);
+	    nabi_log(4, "candidate_format_changed: %s\n",
+			nabi->config->candidate_format);
+	}
+    }
+}
+
 static void
 nabi_candidate_on_row_activated(GtkWidget *widget,
 				GtkTreePath *path,
@@ -286,13 +304,21 @@ nabi_candidate_update_list(NabiCandidate *candidate)
 	const Hanja* hanja = candidate->data[candidate->first + i];
 	const char* value = hanja_get_value(hanja);
 	const char* comment = hanja_get_comment(hanja);
+	char* candidate_str;
+
+	if (nabi_server->use_simplified_chinese) {
+	    candidate_str = nabi_traditional_to_simplified(value);
+	} else {
+	    candidate_str = g_strdup(value);
+	}
 
 	gtk_list_store_append(candidate->store, &iter);
 	gtk_list_store_set(candidate->store, &iter,
 			   COLUMN_INDEX, (i + 1) % 10,
-			   COLUMN_CHARACTER, value,
+			   COLUMN_CHARACTER, candidate_str,
 			   COLUMN_COMMENT, comment,
 			   -1);
+	g_free(candidate_str);
     }
     nabi_candidate_set_window_position(candidate);
 }
@@ -301,6 +327,10 @@ static void
 nabi_candidate_create_window(NabiCandidate *candidate)
 {
     GtkWidget *frame;
+    GtkWidget *vbox;
+    GtkWidget *hbox;
+    GtkWidget *label;
+    GtkWidget *button;
     GtkWidget *treeview;
     GtkTreeViewColumn *column;
     GtkCellRenderer *renderer;
@@ -309,13 +339,42 @@ nabi_candidate_create_window(NabiCandidate *candidate)
 
     nabi_candidate_update_list(candidate);
 
-    frame = gtk_frame_new(candidate->label);
+    frame = gtk_frame_new(NULL);
     gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);
     gtk_container_add(GTK_CONTAINER(candidate->window), frame);
 
+    vbox = gtk_vbox_new(FALSE, 0);
+    gtk_container_add(GTK_CONTAINER(frame), vbox);
+
+    hbox = gtk_hbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);
+
+    label = gtk_label_new(candidate->label);
+    gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 6);
+
+    button = gtk_radio_button_new_with_label(NULL, _("hanja"));
+    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 6);
+    if (strcmp(nabi->config->candidate_format, "hanja") == 0)
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
+    g_signal_connect(G_OBJECT(button), "toggled",
+		     G_CALLBACK(nabi_candidate_on_format), "hanja");
+
+    button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(button), _("hanja(hangul)"));
+    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 6);
+    if (strcmp(nabi->config->candidate_format, "hanja(hangul)") == 0)
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
+    g_signal_connect(G_OBJECT(button), "toggled",
+		     G_CALLBACK(nabi_candidate_on_format), "hanja(hangul)");
+    button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(button), _("hangul(hanja)"));
+    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 6);
+    if (strcmp(nabi->config->candidate_format, "hangul(hanja)") == 0)
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
+    g_signal_connect(G_OBJECT(button), "toggled",
+		     G_CALLBACK(nabi_candidate_on_format), "hangul(hanja)");
+
     treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(candidate->store));
     gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);
-    gtk_container_add(GTK_CONTAINER(frame), treeview);
+    gtk_box_pack_start(GTK_BOX(vbox), treeview, TRUE, TRUE, 0);
     candidate->treeview = treeview;
     g_object_unref(candidate->store);
 
