@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004,2007 Choe Hwanjin
+ * Copyright (C) 2003-2008 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -36,14 +36,17 @@
 #include "server.h"
 #include "fontset.h"
 #include "debug.h"
+#include "util.h"
+#include "nabi.h"
 
 static void nabi_ic_preedit_configure(NabiIC *ic);
 static char* nabi_ic_get_hic_preedit_string(NabiIC *ic);
 static char* nabi_ic_get_preedit_string(NabiIC *ic);
 static char* nabi_ic_get_flush_string(NabiIC *ic);
-static bool  nabi_ic_hic_filter(ucschar* str,
-				ucschar cho, ucschar jung, ucschar jong,
-				void* data);
+static void  nabi_ic_hic_on_translate(HangulInputContext* hic,
+                         int ascii, ucschar* c, void* data);
+static bool  nabi_ic_hic_on_transition(HangulInputContext* hic,
+			 ucschar c, const ucschar* preedit, void* data);
 
 static inline void *
 nabi_malloc(size_t size)
@@ -70,7 +73,7 @@ strniequal(const char* a, const char* b, gsize n)
 }
 
 NabiConnection*
-nabi_connection_create(CARD16 id, const char* encoding)
+nabi_connection_create(CARD16 id, const char* locale)
 {
     NabiConnection* conn;
 
@@ -78,12 +81,21 @@ nabi_connection_create(CARD16 id, const char* encoding)
     conn->id = id;
     conn->mode = NABI_INPUT_MODE_DIRECT;
     conn->cd = (GIConv)-1;
-    if (encoding != NULL) {
-	if (!strniequal(encoding, "UTF-8", 5) ||
-	    !strniequal(encoding, "UTF8", 4)) {
-	    conn->cd = g_iconv_open(encoding, "UTF-8");
+    if (locale != NULL) {
+	char* encoding = strchr(locale, '.');
+	if (encoding != NULL) {
+	    encoding++; // skip '.'
+
+	    if (!strniequal(encoding, "UTF-8", 5) ||
+		!strniequal(encoding, "UTF8", 4)) {
+		conn->cd = g_iconv_open(encoding, "UTF-8");
+		nabi_log(3, "connection %d use encoding: %s (%x)\n",
+			    id, encoding, (int)conn->cd);
+	    }
 	}
     }
+
+    conn->next_new_ic_id = 1;
     conn->ic_list = NULL;
     
     return conn;
@@ -117,6 +129,12 @@ nabi_connection_create_ic(NabiConnection* conn, IMChangeICStruct* data)
 	return NULL;
 
     ic = nabi_ic_create(conn, data);
+    ic->id = conn->next_new_ic_id;
+
+    conn->next_new_ic_id++;
+    if (conn->next_new_ic_id == 0)
+	conn->next_new_ic_id++;
+
     conn->ic_list = g_slist_prepend(conn->ic_list, ic);
     return ic;
 }
@@ -131,6 +149,23 @@ nabi_connection_destroy_ic(NabiConnection* conn, NabiIC* ic)
     nabi_ic_destroy(ic);
 }
 
+NabiIC*
+nabi_connection_get_ic(NabiConnection* conn, CARD16 id)
+{
+    if (conn == NULL || id == 0)
+	return NULL;
+
+    GSList* item = conn->ic_list;
+    while (item != NULL) {
+	NabiIC* ic = (NabiIC*)item->data;
+	if (ic->id == id)
+	    return ic;
+	item = g_slist_next(item);
+    }
+
+    return NULL;
+}
+
 gboolean
 nabi_connection_need_check_charset(NabiConnection* conn)
 {
@@ -160,6 +195,37 @@ nabi_connection_is_valid_str(NabiConnection* conn, const char* str)
     return True;
 }
 
+NabiToplevel*
+nabi_toplevel_new(Window id)
+{
+    NabiToplevel* toplevel = g_new(NabiToplevel, 1);
+
+    toplevel->id = id;
+    toplevel->mode = NABI_INPUT_MODE_DIRECT;
+    toplevel->ref = 1;
+
+    return toplevel;
+}
+
+void
+nabi_toplevel_ref(NabiToplevel* toplevel)
+{
+    if (toplevel != NULL)
+	toplevel->ref++;
+}
+
+void
+nabi_toplevel_unref(NabiToplevel* toplevel)
+{
+    if (toplevel != NULL) {
+	toplevel->ref--;
+	if (toplevel->ref <= 0) {
+	    nabi_server_remove_toplevel(nabi_server, toplevel);
+	    g_free(toplevel);
+	}
+    }
+}
+
 static void
 nabi_ic_init_values(NabiIC *ic)
 {
@@ -224,17 +290,19 @@ nabi_ic_init_values(NabiIC *ic)
 
     ic->candidate = NULL;
 
+    ic->toplevel = NULL;
+
     ic->hic = hangul_ic_new(nabi_server->hangul_keyboard);
-    hangul_ic_set_filter(ic->hic, nabi_ic_hic_filter, ic);
-    ic->has_cho = FALSE;
-    ic->has_jung = FALSE;
-    ic->has_jong = FALSE;
+    hangul_ic_connect_callback(ic->hic, "translate",
+			       nabi_ic_hic_on_translate, ic);
+    hangul_ic_connect_callback(ic->hic, "transition",
+			       nabi_ic_hic_on_transition, ic);
 }
 
 NabiIC*
 nabi_ic_create(NabiConnection* conn, IMChangeICStruct *data)
 {
-    NabiIC *ic = nabi_server_alloc_ic(nabi_server); 
+    NabiIC *ic = g_new(NabiIC, 1);
 
     ic->connection = conn;
 
@@ -250,8 +318,6 @@ nabi_ic_destroy(NabiIC *ic)
     if (ic == NULL)
 	return;
 
-    nabi_ic_preedit_done(ic);
-
     nabi_free(ic->resource_name);
     ic->resource_name = NULL;
     nabi_free(ic->resource_class);
@@ -291,12 +357,17 @@ nabi_ic_destroy(NabiIC *ic)
 	ic->candidate = NULL;
     }
 
+    if (ic->toplevel != NULL) {
+	nabi_toplevel_unref(ic->toplevel);
+	ic->toplevel = NULL;
+    }
+
     if (ic->hic != NULL) {
 	hangul_ic_delete(ic->hic);
 	ic->hic = NULL;
     }
 
-    nabi_server_dealloc_ic(nabi_server, ic);
+    g_free(ic);
 }
 
 CARD16
@@ -326,58 +397,40 @@ nabi_ic_set_hangul_keyboard(NabiIC *ic, const char* hangul_keyboard)
     hangul_ic_select_keyboard(ic->hic, hangul_keyboard);
 }
 
+static void
+nabi_ic_hic_on_translate(HangulInputContext* hic,
+                         int ascii, ucschar* c, void* data)
+{
+    nabi_server_log_key(nabi_server, *c, 0);
+}
+
 static bool
-nabi_ic_hic_filter(ucschar* str, ucschar cho, ucschar jung, ucschar jong,
-	           void* data)
+nabi_ic_hic_on_transition(HangulInputContext* hic,
+			  ucschar c, const ucschar* preedit, void* data)
 {
     bool ret = true;
     NabiIC* ic = (NabiIC*)data;
 
     if (!nabi_server->auto_reorder) {
-	if (ic->has_cho && !ic->has_jung && !ic->has_jong) {
-	    if (cho != 0 && jung == 0 && jong != 0) {
-		ret = false;
-		goto done;
+	if (hangul_is_choseong(c)) {
+	    if (hangul_ic_has_jungseong(hic) || hangul_ic_has_jongseong(hic)) {
+		return false;
 	    }
 	}
 
-	if (!ic->has_cho && ic->has_jung && !ic->has_jong) {
-	    if (cho != 0 && jung != 0 && jong == 0) {
-		ret = false;
-		goto done;
-	    }
-	}
-
-	if (!ic->has_cho && !ic->has_jung && ic->has_jong) {
-	    if (cho != 0 && jung == 0 && jong != 0) {
-		ret = false;
-		goto done;
-	    }
-	    if (cho == 0 && jung != 0 && jong != 0) {
-		ret = false;
-		goto done;
-	    }
-	}
-
-	if (!ic->has_cho && ic->has_jung && ic->has_jong) {
-	    if (cho != 0 && jung != 0 && jong != 0) {
-		ret = false;
-		goto done;
+	if (hangul_is_jungseong(c)) {
+	    if (hangul_ic_has_jongseong(hic)) {
+		return false;
 	    }
 	}
     }
 
     if (ic != NULL) {
-	char* utf8 = g_ucs4_to_utf8((const gunichar*)str, -1, NULL, NULL, NULL);
+	char* utf8 = g_ucs4_to_utf8((const gunichar*)preedit, -1, NULL, NULL, NULL);
 	ret = nabi_connection_is_valid_str(ic->connection, utf8);
 	g_free(utf8);
     }
 
-done:
-    ic->has_cho = cho != 0;
-    ic->has_jung = jung != 0;
-    ic->has_jong = jong != 0;
-
     return ret;
 }
 
@@ -568,7 +621,8 @@ nabi_ic_preedit_hide(NabiIC *ic)
     nabi_log(4, "hide preedit window: id = %d-%d\n",
 	     ic->connection->id, ic->id);
 
-    gdk_window_hide(ic->preedit.window);
+    if (gdk_window_is_visible(ic->preedit.window))
+	gdk_window_hide(ic->preedit.window);
 }
 
 /* move and resize preedit window */
@@ -611,7 +665,9 @@ static GdkFilterReturn
 gdk_event_filter(GdkXEvent *xevent, GdkEvent *gevent, gpointer data)
 {
     XEvent *event = (XEvent*)xevent;
-    NabiIC *ic = nabi_server_get_ic(nabi_server, GPOINTER_TO_UINT(data));
+    guint connect_id = GPOINTER_TO_UINT(data) >> 16;
+    guint ic_id = GPOINTER_TO_UINT(data) & 0xFFFF;
+    NabiIC *ic = nabi_server_get_ic(nabi_server, connect_id, ic_id);
 
     if (ic == NULL)
 	return GDK_FILTER_REMOVE;
@@ -646,6 +702,8 @@ nabi_ic_preedit_window_new(NabiIC *ic)
     GdkWindow *parent = NULL;
     GdkWindowAttr attr;
     gint mask;
+    guint connect_id;
+    guint ic_id;
 
     if (ic->focus_window != 0)
 	parent = gdk_window_foreign_new(ic->focus_window);
@@ -684,11 +742,56 @@ nabi_ic_preedit_window_new(NabiIC *ic)
     gdk_gc_set_background(ic->preedit.hilight_gc, &(nabi_server->preedit_fg));
 
     /* install our preedit window event filter */
+    connect_id = ic->connection->id;
+    ic_id = ic->id;
     gdk_window_add_filter(ic->preedit.window,
-			  gdk_event_filter, GUINT_TO_POINTER((guint)ic->id));
+			  gdk_event_filter,
+			  GUINT_TO_POINTER(connect_id << 16 | ic_id));
     g_object_unref(G_OBJECT(parent));
 }
 
+static void
+nabi_ic_set_client_window(NabiIC* ic, Window client_window)
+{
+    Status s;
+    Window w;
+    Window root = None;
+    Window parent = None;
+    Window* children = NULL;
+    unsigned int  nchildren = 0;
+
+    ic->client_window = client_window;
+    
+    w = client_window;
+    s = XQueryTree(nabi_server->display, w,
+		   &root, &parent, &children, &nchildren);
+    if (s) {
+	while (parent != root) {
+	    if (children != NULL) {
+		XFree(children);
+		children = NULL;
+	    }
+
+	    w = parent;
+	    s = XQueryTree(nabi_server->display, w,
+			   &root, &parent, &children, &nchildren);
+	    if (!s)
+		break;
+	}
+	if (children != NULL) {
+	    XFree(children);
+	    children = NULL;
+	}
+    }
+
+    nabi_log(3, "ic: %d-%d, toplevel: %x\n", ic->id, ic->connection->id, w);
+
+    if (ic->toplevel != NULL)
+	nabi_toplevel_unref(ic->toplevel);
+
+    ic->toplevel = nabi_server_get_toplevel(nabi_server, w);
+}
+
 static void
 nabi_ic_set_focus_window(NabiIC *ic, Window focus_window)
 {
@@ -811,7 +914,7 @@ nabi_ic_set_values(NabiIC *ic, IMChangeICStruct *data)
 	if (streql(XNInputStyle, ic_attr->name)) {
 	    ic->input_style = *(INT32*)ic_attr->value;
 	} else if (streql(XNClientWindow, ic_attr->name)) {
-	    ic->client_window = *(Window*)ic_attr->value;
+	    nabi_ic_set_client_window(ic, *(Window*)ic_attr->value);
 	} else if (streql(XNFocusWindow, ic_attr->name)) {
 	    nabi_ic_set_focus_window(ic, *(Window*)ic_attr->value);
 	} else {
@@ -1032,35 +1135,63 @@ nabi_ic_reset(NabiIC *ic, IMResetICStruct *data)
 }
 
 void
-nabi_ic_mode_direct(NabiIC *ic)
+nabi_ic_set_focus(NabiIC* ic)
 {
-    ic->mode = NABI_INPUT_MODE_DIRECT;
-}
+    NabiInputMode mode = ic->mode;
 
-void
-nabi_ic_mode_compose(NabiIC *ic)
-{
-    ic->mode = NABI_INPUT_MODE_COMPOSE;
+    switch (nabi_server->input_mode_option) {
+    case NABI_INPUT_MODE_PER_DESKTOP:
+	mode = nabi_server->input_mode;
+	break;
+    case NABI_INPUT_MODE_PER_APPLICATION:
+	if (ic->connection != NULL)
+	    mode = ic->connection->mode;
+	break;
+    case NABI_INPUT_MODE_PER_TOPLEVEL:
+	if (ic->toplevel != NULL)
+	    mode = ic->toplevel->mode;
+	break;
+    case NABI_INPUT_MODE_PER_IC:
+    default:
+	break;
+    }
+
+    nabi_ic_set_mode(ic, mode);
+
+    hangul_ic_select_keyboard(ic->hic, nabi_server->hangul_keyboard);
 }
 
 void
 nabi_ic_set_mode(NabiIC *ic, NabiInputMode mode)
 {
+    switch (nabi_server->input_mode_option) {
+    case NABI_INPUT_MODE_PER_DESKTOP:
+	nabi_server->input_mode = mode;
+	break;
+    case NABI_INPUT_MODE_PER_APPLICATION:
+	if (ic->connection != NULL)
+	    ic->connection->mode = mode;
+	break;
+    case NABI_INPUT_MODE_PER_TOPLEVEL:
+	if (ic->toplevel != NULL)
+	    ic->toplevel->mode = mode;
+	break;
+    case NABI_INPUT_MODE_PER_IC:
+    default:
+	break;
+    }
+
     ic->mode = mode;
-    if (ic->connection != NULL)
-	ic->connection->mode = mode;
 
     switch (mode) {
     case NABI_INPUT_MODE_DIRECT:
 	nabi_ic_flush(ic);
 	nabi_ic_preedit_done(ic);
-	if (nabi_server->mode_info_cb != NULL)
-	    nabi_server->mode_info_cb(NABI_MODE_INFO_ENGLISH);
+	nabi_server_set_mode_info(nabi_server, NABI_MODE_INFO_DIRECT);
 	break;
     case NABI_INPUT_MODE_COMPOSE:
 	nabi_ic_preedit_start(ic);
-	if (nabi_server->mode_info_cb != NULL)
-	    nabi_server->mode_info_cb(NABI_MODE_INFO_HANGUL);
+	nabi_server_set_mode_info(nabi_server, NABI_MODE_INFO_COMPOSE);
 	break;
     default:
 	break;
@@ -1367,10 +1498,6 @@ nabi_ic_flush(NabiIC *ic)
     g_free(str);
 
     g_string_erase(ic->preedit.str, 0, -1);
-
-    ic->has_cho = FALSE;
-    ic->has_jung = FALSE;
-    ic->has_jong = FALSE;
 }
 
 void
@@ -1604,7 +1731,10 @@ nabi_ic_process_keyevent(NabiIC* ic, KeySym keysym, unsigned int state)
 	return False;
 
     /* for vi user: on Esc we change state to direct mode */
-    if (keysym == XK_Escape) {
+    if (nabi_server_is_off_key(nabi_server, keysym, state)) {
+	/* 이 경우는 vi 나 emacs등 에디터에서 사용하기 위한 키이므로
+	 * 이 키를 xim에서 사용하지 않고 그대로 다시 forwarding하는 
+	 * 방식으로 작동하게 한다. */
 	nabi_ic_set_mode(ic, NABI_INPUT_MODE_DIRECT);
 	return False;
     }
@@ -1626,6 +1756,9 @@ nabi_ic_process_keyevent(NabiIC* ic, KeySym keysym, unsigned int state)
 	return False;
     }
 
+    /* save key event log */
+    nabi_server_log_key(nabi_server, keysym, state);
+
     if (keysym == XK_BackSpace) {
 	ret = hangul_ic_backspace(ic->hic);
 	if (ret)
@@ -1684,7 +1817,10 @@ nabi_ic_popup_candidate_window (NabiIC *ic)
 	nabi_candidate_delete(ic->candidate);
 
     preedit = nabi_ic_get_preedit_string(ic);
-    list = hanja_table_match_prefix(nabi_server->hanja_table, preedit);
+    list = hanja_table_match_prefix(nabi_server->symbol_table, preedit);
+
+    if (list == NULL)
+	list = hanja_table_match_prefix(nabi_server->hanja_table, preedit);
 
     if (list != NULL) {
 	int i, valid_list_length = 0;
@@ -1710,7 +1846,7 @@ nabi_ic_popup_candidate_window (NabiIC *ic)
 	}
 
 	if (valid_list_length > 0) {
-	    ic->candidate = nabi_candidate_new(NULL, 9,
+	    ic->candidate = nabi_candidate_new(preedit, 9,
 				    list, valid_list, valid_list_length,
 				    parent, &nabi_ic_candidate_commit_cb, ic);
 	    return True;
@@ -1749,8 +1885,32 @@ nabi_ic_insert_candidate(NabiIC *ic, const Hanja* hanja)
 	hangul_ic_reset(ic->hic);
     }
 
-    if (strlen(value) > 0)
-	nabi_ic_commit_utf8(ic, value);
+    if (strlen(value) > 0) {
+	char* modified_value;
+	char* candidate;
+
+	if (nabi_server->use_simplified_chinese) {
+	    modified_value = nabi_traditional_to_simplified(value);
+	    if (!nabi_connection_is_valid_str(ic->connection, modified_value)) {
+		g_free(modified_value);
+		modified_value = g_strdup(value);
+	    }
+	} else {
+	    modified_value = g_strdup(value);
+	}
+
+	if (strcmp(nabi->config->candidate_format, "hanja(hangul)") == 0)
+	    candidate = g_strdup_printf("%s(%s)", modified_value, key);
+	else if (strcmp(nabi->config->candidate_format, "hangul(hanja)") == 0)
+	    candidate = g_strdup_printf("%s(%s)", key, modified_value);
+	else
+	    candidate = g_strdup_printf("%s", modified_value);
+
+	nabi_ic_commit_utf8(ic, candidate);
+
+	g_free(modified_value);
+	g_free(candidate);
+    }
 
     nabi_ic_preedit_update(ic);
 }
