@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004,2007 Choe Hwanjin
+ * Copyright (C) 2003-2008 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,7 +41,7 @@
 #include "fontset.h"
 #include "hangul.h"
 
-#define DEFAULT_IC_TABLE_SIZE	64
+#define NABI_SYMBOL_TABLE NABI_DATA_DIR G_DIR_SEPARATOR_S "symbol.txt"
 
 struct KeySymPair {
     KeySym key;
@@ -63,7 +63,7 @@ Bool nabi_handler(XIMS ims, IMProtocol *call_data);
 
 static void nabi_server_delete_layouts(NabiServer* server);
 
-long nabi_filter_mask = KeyPressMask;
+long nabi_filter_mask = KeyPressMask | KeyReleaseMask;
 
 /* Supported Inputstyles */
 static XIMStyle nabi_input_styles[] = {
@@ -97,7 +97,6 @@ static char *nabi_locales[] = {
 NabiServer*
 nabi_server_new(const char *name)
 {
-    int i;
     NabiServer *server;
 
     server = (NabiServer*)malloc(sizeof(NabiServer));
@@ -110,24 +109,22 @@ nabi_server_new(const char *name)
 
     server->xims = 0;
     server->widget = NULL;
+    server->root_window = NULL;
     server->display = NULL;
     server->window = 0;
     server->filter_mask = 0;
     server->trigger_keys.count_keys = 0;
     server->trigger_keys.keylist = NULL;
+    server->off_keys.count_keys = 0;
+    server->off_keys.keylist = NULL;
     server->candidate_keys.count_keys = 0;
     server->candidate_keys.keylist = NULL;
 
     /* connection list */
     server->connections = NULL;
 
-    /* init IC table */
-    server->last_icid = 0;
-    server->freed_icid = NULL;
-    server->ic_table = g_new(NabiIC*, DEFAULT_IC_TABLE_SIZE);
-    server->ic_table_size = DEFAULT_IC_TABLE_SIZE;
-    for (i = 0; i < server->ic_table_size; i++)
-	server->ic_table[i] = NULL;
+    /* toplevel window list */
+    server->toplevels = NULL;
 
     /* hangul data */
     server->layouts = NULL;
@@ -136,16 +133,20 @@ nabi_server_new(const char *name)
     server->hangul_keyboard_list = hangul_keyboard_list;
 
     server->dynamic_event_flow = True;
-    server->global_input_mode = True;
     server->commit_by_word = False;
-    server->auto_reorder = TRUE;
+    server->auto_reorder = True;
     server->input_mode = NABI_INPUT_MODE_DIRECT;
+    server->input_mode_option = NABI_INPUT_MODE_PER_TOPLEVEL;
 
     /* hanja */
     server->hanja_table = NULL;
 
+    /* symbol */
+    server->symbol_table = NULL;
+
     /* options */
     server->show_status = False;
+    server->use_simplified_chinese = False;
     server->preedit_fg.pixel = 0;
     server->preedit_fg.red = 0xffff;
     server->preedit_fg.green = 0;
@@ -156,9 +157,6 @@ nabi_server_new(const char *name)
     server->preedit_bg.blue = 0;
     server->candidate_font = NULL;
 
-    /* mode info */
-    server->mode_info_cb = NULL;
-
     /* statistics */
     memset(&(server->statistics), 0, sizeof(server->statistics));
 
@@ -168,7 +166,6 @@ nabi_server_new(const char *name)
 void
 nabi_server_destroy(NabiServer *server)
 {
-    int i;
     GSList *item;
 
     if (server == NULL)
@@ -178,23 +175,32 @@ nabi_server_destroy(NabiServer *server)
 	g_object_unref(G_OBJECT(server->gc));
 
     /* destroy remaining connections */
-    item = server->connections;
-    while (item != NULL) {
-	if (item->data != NULL) {
-	    NabiConnection* conn = (NabiConnection*)item->data;
-	    printf("remove connect id: 0x%x\n", conn->id);
-	    nabi_connection_destroy(conn);
+    if (server->connections != NULL) {
+	item = server->connections;
+	while (item != NULL) {
+	    if (item->data != NULL) {
+		NabiConnection* conn = (NabiConnection*)item->data;
+		nabi_log(3, "remove remaining connection: 0x%x\n", conn->id);
+		nabi_connection_destroy(conn);
+	    }
+	    item = g_slist_next(item);
 	}
-	item = g_slist_next(item);
+	g_slist_free(server->connections);
+	server->connections = NULL;
     }
 
-    /* destroy remaining input contexts */
-    for (i = 0; i < server->ic_table_size; i++) {
-	if (server->ic_table[i] != NULL)
-	    nabi_ic_destroy(server->ic_table[i]);
-	server->ic_table[i] = NULL;
+    /* free remaining toplevel list */
+    if (server->toplevels != NULL) {
+	item = server->toplevels;
+	while (item != NULL) {
+	    NabiToplevel* toplevel = (NabiToplevel*)item->data;
+	    nabi_log(3, "remove remaining toplevel: 0x%x\n", toplevel->id);
+	    g_free(item->data);
+	    item = g_slist_next(item);
+	}
+	g_slist_free(server->toplevels);
+	server->toplevels = NULL;
     }
-    g_free(server->ic_table);
 
     /* free remaining fontsets */
     nabi_fontset_free_all(server->display);
@@ -209,6 +215,10 @@ nabi_server_destroy(NabiServer *server)
     if (server->hanja_table != NULL)
 	hanja_table_delete(server->hanja_table);
 
+    /* delete symbol table */
+    if (server->symbol_table != NULL)
+	hanja_table_delete(server->symbol_table);
+
     g_free(server->trigger_keys.keylist);
     g_free(server->candidate_keys.keylist);
     g_free(server->name);
@@ -219,6 +229,9 @@ nabi_server_destroy(NabiServer *server)
 void
 nabi_server_set_hangul_keyboard(NabiServer *server, const char *id)
 {
+    if (server == NULL)
+	return;
+
     if (server->hangul_keyboard != NULL)
 	g_free(server->hangul_keyboard);
 
@@ -226,12 +239,19 @@ nabi_server_set_hangul_keyboard(NabiServer *server, const char *id)
 }
 
 void
-nabi_server_set_mode_info_cb(NabiServer *server, NabiModeInfoCallback func)
+nabi_server_set_mode_info(NabiServer *server, int state)
 {
+    guint32 data;
+
     if (server == NULL)
 	return;
 
-    server->mode_info_cb = func;
+    data = state;
+    gdk_property_change(server->root_window,
+			gdk_atom_intern ("_HANGUL_INPUT_MODE", FALSE),
+			gdk_atom_intern ("INTEGER", FALSE),
+			32, GDK_PROP_MODE_REPLACE,
+			(const guchar *)&data, 1);
 }
 
 void
@@ -259,12 +279,27 @@ nabi_server_set_candidate_font(NabiServer *server, const gchar *font_name)
     }
 }
 
-static XIMTriggerKey*
-xim_trigger_keys_create(char **keys, int n)
+static void
+xim_trigger_keys_set_value(XIMTriggerKeys* keys, char** key_strings)
 {
-    int i, j;
+    int i, j, n;
     XIMTriggerKey *keylist;
 
+    if (keys == NULL)
+	return;
+
+    if (keys->keylist != NULL) {
+	g_free(keys->keylist);
+	keys->keylist = NULL;
+	keys->count_keys = 0;
+    }
+
+    if (key_strings == NULL)
+	return;
+
+    for (n = 0; key_strings[n] != NULL; n++)
+	continue;
+
     keylist = g_new(XIMTriggerKey, n);
     for (i = 0; i < n; i++) {
 	keylist[i].keysym = 0;
@@ -273,7 +308,7 @@ xim_trigger_keys_create(char **keys, int n)
     }
 
     for (i = 0; i < n; i++) {
-	gchar **list = g_strsplit(keys[i], "+", 0);
+	gchar **list = g_strsplit(key_strings[i], "+", 0);
 
 	for (j = 0; list[j] != NULL; j++) {
 	    if (strcmp("Shift", list[j]) == 0) {
@@ -294,50 +329,40 @@ xim_trigger_keys_create(char **keys, int n)
 	}
     }
 
-    return keylist;
+    keys->keylist = keylist;
+    keys->count_keys = n;
 }
 
 void
 nabi_server_set_trigger_keys(NabiServer *server, char **keys)
 {
-    int n;
+    xim_trigger_keys_set_value(&server->trigger_keys, keys);
 
-    if (keys == NULL)
-	return;
+    if (server->xims != NULL && server->dynamic_event_flow) {
+	IMSetIMValues(server->xims,
+		      IMOnKeysList, &(server->trigger_keys),
+		      NULL);
+    }
+}
 
-    if (server->trigger_keys.keylist != NULL)
-	g_free(server->trigger_keys.keylist);
-
-    for (n = 0; keys[n] != NULL; n++)
-	continue;
-
-    server->trigger_keys.keylist = xim_trigger_keys_create(keys, n);
-    server->trigger_keys.count_keys = n;
+void
+nabi_server_set_off_keys(NabiServer *server, char **keys)
+{
+    xim_trigger_keys_set_value(&server->off_keys, keys);
 }
 
 void
 nabi_server_set_candidate_keys(NabiServer *server, char **keys)
 {
-    int n;
-
-    if (keys == NULL)
-	return;
-
-    if (server->candidate_keys.keylist != NULL)
-	g_free(server->candidate_keys.keylist);
-
-    for (n = 0; keys[n] != NULL; n++)
-	continue;
-
-    server->candidate_keys.keylist = xim_trigger_keys_create(keys, n);
-    server->candidate_keys.count_keys = n;
+    xim_trigger_keys_set_value(&server->candidate_keys, keys);
 }
 
 void
 nabi_server_init(NabiServer *server)
 {
     const char *charset;
-    char *trigger_keys[3] = { "Hangul", "Shift+space", NULL };
+    char *trigger_keys[3]   = { "Hangul", "Shift+space", NULL };
+    char *off_keys[2]       = { "Escape", NULL };
     char *candidate_keys[3] = { "Hangul_Hanja", "F9", NULL };
 
     if (server == NULL)
@@ -347,6 +372,7 @@ nabi_server_init(NabiServer *server)
     server->filter_mask = KeyPressMask;
 
     nabi_server_set_trigger_keys(server, trigger_keys);
+    nabi_server_set_off_keys(server, off_keys);
     nabi_server_set_candidate_keys(server, candidate_keys);
 
     server->output_mode = NABI_OUTPUT_SYLLABLE;
@@ -364,71 +390,35 @@ nabi_server_init(NabiServer *server)
     }
 
     server->hanja_table = hanja_table_load(NULL);
-}
-
-NabiIC*
-nabi_server_alloc_ic(NabiServer* server)
-{
-    CARD16 icid = 0;
-    NabiIC* ic;
-
-    if (server->freed_icid == NULL) {
-	server->last_icid++;
-	if (server->last_icid == 0)
-	    server->last_icid++;  /* icid에 0은 사용하지 않는다 */
-
-	if (server->last_icid >= server->ic_table_size) {
-	    gint new_size = server->ic_table_size;
-	    while (new_size <= server->last_icid) {
-		new_size += DEFAULT_IC_TABLE_SIZE / 2;
-	    }
-	    server->ic_table = g_renew(NabiIC*, server->ic_table, new_size);
-	}
-
-	icid = server->last_icid;
-    } else {
-	/* pick from ic_freed */
-	icid = (CARD16)GPOINTER_TO_UINT(server->freed_icid->data);
-	server->freed_icid = g_slist_remove(server->freed_icid, 
-					    GUINT_TO_POINTER((guint)icid));
-    }
-
-    ic = g_new(NabiIC, 1);
-    ic->id = icid;
-
-    server->ic_table[ic->id] = ic;
-
-    return ic;
-}
-
-void
-nabi_server_dealloc_ic(NabiServer* server, NabiIC* ic)
-{
-    server->freed_icid = g_slist_append(server->freed_icid,
-					GUINT_TO_POINTER((guint)ic->id));
-    server->ic_table[ic->id] = NULL;
-
-    g_free(ic);
+    server->symbol_table = hanja_table_load(NABI_SYMBOL_TABLE);
 }
 
 gboolean
 nabi_server_is_valid_ic(NabiServer* server, NabiIC* ic)
 {
-    int i;
-    for (i = 0; i < server->ic_table_size; i++) {
-	if (server->ic_table[i] == ic)
+    GSList* item = server->connections;
+
+    while (item != NULL) {
+	NabiConnection* conn = (NabiConnection*)item->data;
+	if (g_slist_find(conn->ic_list, ic) != NULL)
 	    return TRUE;
+	item = g_slist_next(item);
     }
+
     return FALSE;
 }
 
 NabiIC*
-nabi_server_get_ic(NabiServer *server, CARD16 icid)
+nabi_server_get_ic(NabiServer *server, CARD16 connect_id, CARD16 ic_id)
 {
-    if (icid > 0 && icid < server->ic_table_size)
-	return server->ic_table[icid];
-    else
-	return NULL;
+    NabiConnection* conn;
+
+    conn = nabi_server_get_connection(server, connect_id);
+    if (conn != NULL) {
+	return nabi_connection_get_ic(conn, ic_id);
+    }
+
+    return NULL;
 }
 
 static Bool
@@ -453,6 +443,14 @@ nabi_server_is_trigger_key(NabiServer* server, KeySym key, unsigned int state)
 			      key, state);
 }
 
+Bool
+nabi_server_is_off_key(NabiServer* server, KeySym key, unsigned int state)
+{
+    return nabi_server_is_key(server->off_keys.keylist,
+			      server->off_keys.count_keys,
+			      key, state);
+}
+
 Bool
 nabi_server_is_candidate_key(NabiServer* server, KeySym key, unsigned int state)
 {
@@ -461,6 +459,28 @@ nabi_server_is_candidate_key(NabiServer* server, KeySym key, unsigned int state)
 			      key, state);
 }
 
+Bool
+nabi_server_is_running(const char* name)
+{
+    Display* display;
+    Atom atom;
+    Window owner;
+    char atom_name[64];
+
+    display = gdk_x11_get_default_xdisplay();
+
+    snprintf(atom_name, sizeof(atom_name), "@server=%s", name);
+
+    atom = XInternAtom(display, atom_name, True);
+    if (atom != None) { 
+	owner = XGetSelectionOwner(display, atom);
+	if (owner != None)
+	    return True;
+    }
+
+    return False;
+}
+
 int
 nabi_server_start(NabiServer *server, GtkWidget *widget)
 {
@@ -500,7 +520,7 @@ nabi_server_start(NabiServer *server, GtkWidget *widget)
     g_free(locales);
 
     if (xims == NULL) {
-	fprintf(stderr, "Nabi: Can't open Input Method Service\n");
+	nabi_log(1, "can't open input method service\n");
 	exit(1);
     }
 
@@ -518,6 +538,7 @@ nabi_server_start(NabiServer *server, GtkWidget *widget)
 
     server->xims = xims;
     server->widget = widget;
+    server->root_window = gdk_screen_get_root_window(gtk_widget_get_screen(widget));
     server->display = display;
     server->window = window;
 
@@ -525,6 +546,8 @@ nabi_server_start(NabiServer *server, GtkWidget *widget)
     gdk_gc_set_foreground(server->gc, &(server->preedit_fg));
     gdk_gc_set_background(server->gc, &(server->preedit_bg));
 
+    server->start_time = time(NULL);
+
     nabi_log(1, "xim server started\n");
 
     return 0;
@@ -550,18 +573,11 @@ nabi_server_create_connection(NabiServer *server,
 			      CARD16 connect_id, const char* locale)
 {
     NabiConnection* conn;
-    const char* encoding = NULL;
 
     if (server == NULL)
 	return NULL;
 
-    if (locale != NULL) {
-	encoding = strchr(locale, '.');
-	if (encoding != NULL)
-	    encoding++; // skip '.'
-    }
-
-    conn = nabi_connection_create(connect_id, encoding);
+    conn = nabi_connection_create(connect_id, locale);
     server->connections = g_slist_prepend(server->connections, conn);
     return conn;
 }
@@ -591,6 +607,35 @@ nabi_server_destroy_connection(NabiServer *server, CARD16 connect_id)
     nabi_connection_destroy(conn);
 }
 
+NabiToplevel*
+nabi_server_get_toplevel(NabiServer* server, Window id)
+{
+    NabiToplevel* toplevel;
+    GSList* item;
+
+    item = server->toplevels;
+    while (item != NULL) {
+	NabiToplevel* toplevel = (NabiToplevel*)item->data;
+	if (toplevel != NULL && toplevel->id == id) {
+	    nabi_toplevel_ref(toplevel);
+	    return toplevel;
+	}
+	item = g_slist_next(item);
+    }
+
+    toplevel = nabi_toplevel_new(id);
+    server->toplevels = g_slist_prepend(server->toplevels, toplevel);
+
+    return toplevel;
+}
+
+void
+nabi_server_remove_toplevel(NabiServer* server, NabiToplevel* toplevel)
+{
+    if (server != NULL && server->toplevels != NULL)
+	server->toplevels = g_slist_remove(server->toplevels, toplevel);
+}
+
 Bool
 nabi_server_is_locale_supported(NabiServer *server, const char *locale)
 {
@@ -607,22 +652,22 @@ nabi_server_is_locale_supported(NabiServer *server, const char *locale)
 }
 
 void
-nabi_server_on_keypress(NabiServer *server,
-			KeySym keyval,
-			unsigned int state,
-			wchar_t ch)
+nabi_server_log_key(NabiServer *server, ucschar c, unsigned int state)
 {
-    server->statistics.total++;
-
-    if (keyval == XK_BackSpace)
+    if (c == XK_BackSpace) {
 	server->statistics.backspace++;
-    else if (keyval == XK_space)
+	server->statistics.total++;
+    } else if (c == XK_space) {
 	server->statistics.space++;
-    else if (ch >= 0x1100 && ch <= 0x11FF) {
-	int index = (unsigned int)ch & 0xff;
-	if (index >= 0 && index <= 255)
+	server->statistics.total++;
+    } else if (c >= 0x1100 && c <= 0x11FF) {
+	int index = (unsigned int)c & 0xff;
+	if (index >= 0 && index <= 255) {
 	    server->statistics.jamo[index]++;
+	    server->statistics.total++;
+	}
     }
+
     if (state & ShiftMask)
 	server->statistics.shift++;
 }
@@ -772,6 +817,9 @@ nabi_server_set_keyboard_layout(NabiServer *server, const char* name)
 {
     GList* list;
 
+    if (server != NULL)
+	return;
+
     if (strcmp(name, "none") == 0) {
 	server->layout = NULL;
 	return;
@@ -790,7 +838,27 @@ nabi_server_set_keyboard_layout(NabiServer *server, const char* name)
 const NabiHangulKeyboard*
 nabi_server_get_hangul_keyboard_list(NabiServer* server)
 {
-    return server->hangul_keyboard_list;
+    if (server != NULL)
+	return server->hangul_keyboard_list;
+
+    return NULL;
+}
+
+const char*
+nabi_server_get_keyboard_name_by_id(NabiServer* server, const char* id)
+{
+    int i;
+
+    if (server == NULL)
+	return NULL;
+
+    for (i = 0; server->hangul_keyboard_list[i].id != NULL; i++) {
+	if (strcmp(id, server->hangul_keyboard_list[i].id) == 0) {
+	    return server->hangul_keyboard_list[i].name;
+	}
+    }
+
+    return NULL;
 }
 
 KeySym
@@ -820,29 +888,67 @@ nabi_server_normalize_keysym(NabiServer *server,
     return keysym;
 }
 
+void
+nabi_server_toggle_input_mode(NabiServer* server)
+{
+    if (server == NULL)
+	return;
+
+    if (server->input_mode_option == NABI_INPUT_MODE_PER_DESKTOP) {
+	if (server->input_mode == NABI_INPUT_MODE_DIRECT) {
+	    server->input_mode = NABI_INPUT_MODE_COMPOSE;
+	    nabi_server_set_mode_info(nabi_server, NABI_MODE_INFO_COMPOSE);
+	    nabi_log(1, "change input mode: compose\n");
+	} else {
+	    server->input_mode = NABI_INPUT_MODE_DIRECT;
+	    nabi_server_set_mode_info(nabi_server, NABI_MODE_INFO_DIRECT);
+	    nabi_log(1, "change input mode: direct\n");
+	}
+    }
+}
+
 void
 nabi_server_set_dynamic_event_flow(NabiServer* server, Bool flag)
 {
-    server->dynamic_event_flow = flag;
+    if (server != NULL)
+	server->dynamic_event_flow = flag;
 }
 
 void
 nabi_server_set_xim_name(NabiServer* server, const char* name)
 {
-    g_free(server->name);
-    server->name = g_strdup(name);
+    if (server != NULL) {
+	g_free(server->name);
+	server->name = g_strdup(name);
+    }
 }
 
 void
 nabi_server_set_commit_by_word(NabiServer* server, Bool flag)
 {
-    server->commit_by_word = flag;
+    if (server != NULL)
+	server->commit_by_word = flag;
 }
 
 void
 nabi_server_set_auto_reorder(NabiServer* server, Bool flag)
 {
-    server->auto_reorder = flag;
+    if (server != NULL)
+	server->auto_reorder = flag;
+}
+
+void
+nabi_server_set_input_mode_option(NabiServer* server, NabiInputModeOption flag)
+{
+    if (server != NULL)
+	server->input_mode_option = flag;
+}
+
+void
+nabi_server_set_simplified_chinese(NabiServer* server, Bool state)
+{
+    if (server != NULL)
+	server->use_simplified_chinese = state;
 }
 
 void
@@ -850,165 +956,88 @@ nabi_server_write_log(NabiServer *server)
 {
     const gchar *homedir;
     gchar *filename;
-    time_t current_time;
-    struct tm current_tm;
-    char date[256] = { '\0', };
     FILE *file;
 
+    if (server->statistics.total <= 0)
+	return;
+
     homedir = g_get_home_dir();
     filename = g_build_filename(homedir, ".nabi", "nabi.log", NULL);
     
     file = fopen(filename, "a");
     if (file != NULL) {
+	int i, sum;
+	time_t current_time;
+	struct tm local_time;
+	char buf[256] = { '\0', };
+
+	localtime_r(&server->start_time, &local_time);
+	strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", &local_time);
+	fprintf(file, "%s - ", buf);
+
 	current_time = time(NULL);
-	localtime_r(&current_time, &current_tm);
-	strftime(date, sizeof(date), "%Y-%m-%d %H:%M:%S", &current_tm);
+	localtime_r(&current_time, &local_time);
+	strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", &local_time);
+
+	fprintf(file, "%s", buf);
+	fprintf(file, " (%ds)\n", (int)(current_time - server->start_time));
 
-	fprintf(file, "%s\n", date);
 	fprintf(file, "total: %d\n", server->statistics.total);
 	fprintf(file, "space: %d\n", server->statistics.space);
 	fprintf(file, "backspace: %d\n", server->statistics.backspace);
 	fprintf(file, "keyboard: %s\n", server->hangul_keyboard);
 
-	fprintf(file, "cho:  "
-		      "\341\204\200: %-3d "
-		      "\341\204\201: %-3d "
-		      "\341\204\202: %-3d "
-		      "\341\204\203: %-3d "
-		      "\341\204\204: %-3d "
-		      "\341\204\205: %-3d "
-		      "\341\204\206: %-3d "
-		      "\341\204\207: %-3d "
-		      "\341\204\210: %-3d "
-		      "\341\204\211: %-3d "
-		      "\341\204\212: %-3d "
-		      "\341\204\213: %-3d "
-		      "\341\204\214: %-3d "
-		      "\341\204\215: %-3d "
-		      "\341\204\216: %-3d "
-		      "\341\204\217: %-3d "
-		      "\341\204\220: %-3d "
-		      "\341\204\221: %-3d "
-		      "\341\204\222: %-3d\n",
-		      nabi_server->statistics.jamo[0x00],
-		      nabi_server->statistics.jamo[0x01],
-		      nabi_server->statistics.jamo[0x02],
-		      nabi_server->statistics.jamo[0x03],
-		      nabi_server->statistics.jamo[0x04],
-		      nabi_server->statistics.jamo[0x05],
-		      nabi_server->statistics.jamo[0x06],
-		      nabi_server->statistics.jamo[0x07],
-		      nabi_server->statistics.jamo[0x08],
-		      nabi_server->statistics.jamo[0x09],
-		      nabi_server->statistics.jamo[0x0a],
-		      nabi_server->statistics.jamo[0x0b],
-		      nabi_server->statistics.jamo[0x0c],
-		      nabi_server->statistics.jamo[0x0d],
-		      nabi_server->statistics.jamo[0x0e],
-		      nabi_server->statistics.jamo[0x0f],
-		      nabi_server->statistics.jamo[0x10],
-		      nabi_server->statistics.jamo[0x11],
-		      nabi_server->statistics.jamo[0x12]);
+	/* choseong */
+	sum = 0; 
+	for (i = 0x00; i <= 0x12; i++)
+	    sum += server->statistics.jamo[i];
 
-	fprintf(file, "jung: "
-		      "\341\205\241: %-3d "
-		      "\341\205\242: %-3d "
-		      "\341\205\243: %-3d "
-		      "\341\205\244: %-3d "
-		      "\341\205\245: %-3d "
-		      "\341\205\246: %-3d "
-		      "\341\205\247: %-3d "
-		      "\341\205\250: %-3d "
-		      "\341\205\251: %-3d "
-		      "\341\205\252: %-3d "
-		      "\341\205\253: %-3d "
-		      "\341\205\254: %-3d "
-		      "\341\205\255: %-3d "
-		      "\341\205\256: %-3d "
-		      "\341\205\257: %-3d "
-		      "\341\205\260: %-3d "
-		      "\341\205\261: %-3d "
-		      "\341\205\262: %-3d "
-		      "\341\205\263: %-3d "
-		      "\341\205\264: %-3d "
-		      "\341\205\265: %-3d\n",
-		      nabi_server->statistics.jamo[0x61],
-		      nabi_server->statistics.jamo[0x62],
-		      nabi_server->statistics.jamo[0x63],
-		      nabi_server->statistics.jamo[0x64],
-		      nabi_server->statistics.jamo[0x65],
-		      nabi_server->statistics.jamo[0x66],
-		      nabi_server->statistics.jamo[0x67],
-		      nabi_server->statistics.jamo[0x68],
-		      nabi_server->statistics.jamo[0x69],
-		      nabi_server->statistics.jamo[0x6a],
-		      nabi_server->statistics.jamo[0x6b],
-		      nabi_server->statistics.jamo[0x6c],
-		      nabi_server->statistics.jamo[0x6d],
-		      nabi_server->statistics.jamo[0x6e],
-		      nabi_server->statistics.jamo[0x6f],
-		      nabi_server->statistics.jamo[0x70],
-		      nabi_server->statistics.jamo[0x71],
-		      nabi_server->statistics.jamo[0x72],
-		      nabi_server->statistics.jamo[0x73],
-		      nabi_server->statistics.jamo[0x74],
-		      nabi_server->statistics.jamo[0x75]);
+	if (sum > 0) {
+	    fprintf(file, "cho: ");
+	    for (i = 0x0; i <= 0x12; i++) {
+		char label[8] = { '\0', };
+		g_unichar_to_utf8(0x1100 + i, label);
+		fprintf(file, "%s: %-3d ",
+			label, server->statistics.jamo[i]);
+	    }
+	    fprintf(file, "\n");
+	} else {
+	    fprintf(file, "cho: 0\n");
+	}
 
-	fprintf(file, "jong: "
-		      "\341\206\250: %-3d "
-		      "\341\206\251: %-3d "
-		      "\341\206\252: %-3d "
-		      "\341\206\253: %-3d "
-		      "\341\206\254: %-3d "
-		      "\341\206\255: %-3d "
-		      "\341\206\256: %-3d "
-		      "\341\206\257: %-3d "
-		      "\341\206\260: %-3d "
-		      "\341\206\261: %-3d "
-		      "\341\206\262: %-3d "
-		      "\341\206\263: %-3d "
-		      "\341\206\264: %-3d "
-		      "\341\206\265: %-3d "
-		      "\341\206\266: %-3d "
-		      "\341\206\267: %-3d "
-		      "\341\206\270: %-3d "
-		      "\341\206\271: %-3d "
-		      "\341\206\272: %-3d "
-		      "\341\206\273: %-3d "
-		      "\341\206\274: %-3d "
-		      "\341\206\275: %-3d "
-		      "\341\206\276: %-3d "
-		      "\341\206\277: %-3d "
-		      "\341\207\200: %-3d "
-		      "\341\207\201: %-3d "
-		      "\341\207\202: %-3d\n",
-		      nabi_server->statistics.jamo[0xa8],
-		      nabi_server->statistics.jamo[0xa9],
-		      nabi_server->statistics.jamo[0xaa],
-		      nabi_server->statistics.jamo[0xab],
-		      nabi_server->statistics.jamo[0xac],
-		      nabi_server->statistics.jamo[0xad],
-		      nabi_server->statistics.jamo[0xae],
-		      nabi_server->statistics.jamo[0xaf],
-		      nabi_server->statistics.jamo[0xb0],
-		      nabi_server->statistics.jamo[0xb1],
-		      nabi_server->statistics.jamo[0xb2],
-		      nabi_server->statistics.jamo[0xb3],
-		      nabi_server->statistics.jamo[0xb4],
-		      nabi_server->statistics.jamo[0xb5],
-		      nabi_server->statistics.jamo[0xb6],
-		      nabi_server->statistics.jamo[0xb7],
-		      nabi_server->statistics.jamo[0xb8],
-		      nabi_server->statistics.jamo[0xb9],
-		      nabi_server->statistics.jamo[0xba],
-		      nabi_server->statistics.jamo[0xbb],
-		      nabi_server->statistics.jamo[0xbc],
-		      nabi_server->statistics.jamo[0xbd],
-		      nabi_server->statistics.jamo[0xbe],
-		      nabi_server->statistics.jamo[0xbf],
-		      nabi_server->statistics.jamo[0xc0],
-		      nabi_server->statistics.jamo[0xc1],
-		      nabi_server->statistics.jamo[0xc2]);
+	/* jungseong */
+	sum = 0; 
+	for (i = 0x61; i <= 0x75; i++)
+	    sum += server->statistics.jamo[i];
+
+	if (sum > 0) {
+	    fprintf(file, "jung: ");
+	    for (i = 0x61; i <= 0x75; i++) {
+		char label[8] = { '\0', };
+		g_unichar_to_utf8(0x1100 + i, label);
+		fprintf(file, "%s: %-3d ",
+			label, server->statistics.jamo[i]);
+	    }
+	    fprintf(file, "\n");
+	} else {
+	    fprintf(file, "jung: 0\n");
+	}
+
+	/* jong seong */
+	sum = 0; 
+	for (i = 0xa8; i <= 0xc2; i++)
+	    sum += server->statistics.jamo[i];
+
+	if (sum > 0) {
+	    fprintf(file, "jong: ");
+	    for (i = 0xa8; i <= 0xc2; i++) {
+		char label[8] = { '\0', };
+		g_unichar_to_utf8(0x1100 + i, label);
+		fprintf(file, "%s: %-3d ",
+			label, server->statistics.jamo[i]);
+	    }
+	    fprintf(file, "\n");
+	}
 
 	fprintf(file, "\n");
 
