@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004,2007 Choe Hwanjin
+ * Copyright (C) 2003-2008 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,6 +25,8 @@
 
 #include <X11/Xlib.h>
 #include <wchar.h>
+#include <time.h>
+
 #include <glib.h>
 #include <gtk/gtk.h>
 
@@ -50,18 +52,6 @@ struct _NabiKeyboardLayout {
     GArray* table;
 };
 
-enum {
-    NABI_TRIGGER_KEY_HANGUL      = 1 << 0,
-    NABI_TRIGGER_KEY_SHIFT_SPACE = 1 << 1,
-    NABI_TRIGGER_KEY_ALT_R       = 1 << 2
-};
-
-enum {
-    NABI_CANDIDATE_KEY_HANJA  = 1 << 0,
-    NABI_CANDIDATE_KEY_F9     = 1 << 1,
-    NABI_CANDIDATE_KEY_CTRL_R = 1 << 2
-};
-
 typedef enum {
     NABI_OUTPUT_SYLLABLE,
     NABI_OUTPUT_JAMO,
@@ -69,14 +59,16 @@ typedef enum {
 } NabiOutputMode;
 
 typedef enum {
-    NABI_KEYBOARD_2SET,
-    NABI_KEYBOARD_3SET
-} NabiKeyboardType;
+    NABI_INPUT_MODE_PER_DESKTOP,
+    NABI_INPUT_MODE_PER_APPLICATION,
+    NABI_INPUT_MODE_PER_TOPLEVEL,
+    NABI_INPUT_MODE_PER_IC
+} NabiInputModeOption;
 
 enum {
     NABI_MODE_INFO_NONE,
-    NABI_MODE_INFO_ENGLISH,
-    NABI_MODE_INFO_HANGUL
+    NABI_MODE_INFO_DIRECT,
+    NABI_MODE_INFO_COMPOSE
 };
 
 typedef void (*NabiModeInfoCallback)(int);
@@ -94,22 +86,19 @@ struct _NabiServer {
     char*		    name;
     XIMS                    xims;
     GtkWidget*              widget;
+    GdkWindow*              root_window;
     Display*                display;
     Window                  window;
     long                    filter_mask;
     XIMTriggerKeys          trigger_keys;
+    XIMTriggerKeys          off_keys;
     XIMTriggerKeys          candidate_keys;
     char**                  locales;
     GdkGC*		    gc;
 
-    /* xim connect list */
+    /* xim connection list */
     GSList*                 connections;
-
-    /* Input Context list */
-    CARD16                  last_icid;
-    GSList*                 freed_icid;
-    NabiIC**                ic_table;
-    int                     ic_table_size;
+    GSList*                 toplevels;
 
     /* keyboard translate */
     GList*                  layouts;
@@ -124,21 +113,23 @@ struct _NabiServer {
     /* hanja */
     HanjaTable*             hanja_table;
 
+    /* symbol */
+    HanjaTable*             symbol_table;
+
     /* options */
     Bool                    dynamic_event_flow;
-    Bool                    global_input_mode;
     Bool                    commit_by_word;
     Bool                    auto_reorder;
     Bool                    show_status;
+    Bool                    use_simplified_chinese;
     NabiInputMode           input_mode;
+    NabiInputModeOption     input_mode_option;
     GdkColor                preedit_fg;
     GdkColor                preedit_bg;
     gchar*		    candidate_font;
 
-    /* mode information */
-    NabiModeInfoCallback    mode_info_cb;
-
     /* statistics */
+    time_t                  start_time;
     struct NabiStatistics   statistics;
 };
 
@@ -151,14 +142,18 @@ int         nabi_server_start           (NabiServer* server,
 					 GtkWidget*  widget);
 int         nabi_server_stop            (NabiServer *server);
 
+Bool        nabi_server_is_running();
 Bool        nabi_server_is_trigger_key  (NabiServer*  server,
                                          KeySym       key,
                                          unsigned int state);
+Bool        nabi_server_is_off_key      (NabiServer*  server,
+                                         KeySym       key,
+                                         unsigned int state);
 Bool        nabi_server_is_candidate_key(NabiServer*  server,
                                          KeySym       key,
                                          unsigned int state);
-void        nabi_server_set_trigger_keys(NabiServer *server,
-					 char **keys);
+void        nabi_server_set_trigger_keys  (NabiServer *server, char **keys);
+void        nabi_server_set_off_keys      (NabiServer *server, char **keys);
 void        nabi_server_set_candidate_keys(NabiServer *server, char **keys);
 
 void        nabi_server_load_keyboard_layout(NabiServer *server,
@@ -167,9 +162,9 @@ void        nabi_server_set_keyboard_layout(NabiServer *server,
 					    const char* name);
 void        nabi_server_set_hangul_keyboard(NabiServer *server,
 					 const char *id);
+void        nabi_server_toggle_input_mode(NabiServer* server);
 
-void        nabi_server_set_mode_info_cb(NabiServer *server,
-					 NabiModeInfoCallback func);
+void        nabi_server_set_mode_info(NabiServer *server, int state);
 void        nabi_server_set_output_mode (NabiServer *server,
 					 NabiOutputMode mode);
 void	    nabi_server_set_candidate_font(NabiServer *server,
@@ -178,10 +173,12 @@ void        nabi_server_set_dynamic_event_flow(NabiServer* server, Bool flag);
 void        nabi_server_set_xim_name(NabiServer* server, const char* name);
 void        nabi_server_set_commit_by_word(NabiServer* server, Bool flag);
 void        nabi_server_set_auto_reorder(NabiServer* server, Bool flag);
+void        nabi_server_set_input_mode_option(NabiServer* server,
+					      NabiInputModeOption flag);
+void        nabi_server_set_simplified_chinese(NabiServer* server, Bool state);
 
-NabiIC*     nabi_server_alloc_ic        (NabiServer* server);
-void        nabi_server_dealloc_ic      (NabiServer* server, NabiIC* ic);
-NabiIC*     nabi_server_get_ic          (NabiServer *server, CARD16 icid);
+NabiIC*     nabi_server_get_ic          (NabiServer *server,
+					 CARD16 connect_id, CARD16 icid);
 gboolean    nabi_server_is_valid_ic     (NabiServer* server, NabiIC* ic);
 
 NabiConnection* nabi_server_create_connection (NabiServer *server,
@@ -192,15 +189,18 @@ NabiConnection* nabi_server_get_connection    (NabiServer *server,
 void            nabi_server_destroy_connection(NabiServer *server,
 					       CARD16 connect_id);
 
+NabiToplevel*   nabi_server_get_toplevel(NabiServer* server, Window id);
+void            nabi_server_remove_toplevel(NabiServer* server,
+					    NabiToplevel* toplevel);
+
 Bool        nabi_server_is_locale_supported(NabiServer *server,
 					    const char *locale);
 Bool        nabi_server_is_valid_str    (NabiServer *server, const char* str);
 KeySym      nabi_server_normalize_keysym(NabiServer *server,
 					 KeySym keysym, unsigned int state);
-void        nabi_server_on_keypress     (NabiServer *server,
-					 KeySym keyval,
-					 unsigned int state,
-					 wchar_t ch);
+void        nabi_server_log_key         (NabiServer *server,
+					 ucschar c,
+					 unsigned int state);
 void        nabi_server_write_log(NabiServer *server);
 
 Bool	    nabi_server_load_keyboard_table(NabiServer *server,
@@ -211,6 +211,9 @@ Bool	    nabi_server_load_compose_table(NabiServer *server,
 const NabiHangulKeyboard*
 nabi_server_get_hangul_keyboard_list(NabiServer* server);
 
+const char*
+nabi_server_get_keyboard_name_by_id(NabiServer* server, const char* id);
+
 #endif  /* __SERVER_H_ */
 
 /* vim: set ts=8 sw=4 sts=4 : */
